/* Arduino (nano) programma voor de sector draaischijf in  
 *  het Himbeeren Wald schaduwstation
 *  
 *  (c) Fenna Pel, februari 2020, maart 2020
 *  pereltje@pereltje.nl
 *  
 *  Deze code is "as is" beschikbaar gesteld onder de GPL v3 licentie
 *  zonder garantie voor geschiktheid voor enig doel
 *  
 *  sector draaischijf Himbeeren Wald modelspoor project:
 *  7 aangesloten sporen, spoor 1 sluit aan naar de hoofdbaan
 *  sporen 2, 3, 4 en 5 hebben elk 2 opstelmogelijkheden voor complete treinen
 *  sporen 6 en 7 hebben elk ruimte voor 1 tractie voertuig.
 *  
 *  Het maximum aantal sporen is in de code verwerkt met de constante 'aantalsporen'
 *  Door deze waarde te veranderen en op de spoor pinnen binaire decoders aan te sluiten,
 *  kan het aantal sporen worden uitgebreid.
 *  
 *  Het gebruik van de beschikbare pinnen staat vermeld aan het begin van de code (const defenities)
 *  
 *  linkspin en rechtspin zijn bedoeld voor hand bediening met drukknoppen (aan massa leggen)
 *  
 *  pin 13 is intern in de Nano gekoppeld aan een on-board led en wordt gebruikt voor een 'bezig' signaal
 *  dat evt met de RPi wordt gedeeld
 *  
 *  pin 0 en 1 worden gedeeld met de usb verbinding. Niet gebruiken voor andere toepassingen zolang usb 
 *  verbinding met een host wenselijk is.
 *  
 *  Deze code voorziet niet in het omgaan met een draaischijf die volledig rond kan draaien. 
 *  Diegenen die daar behoefte aan hebben, zijn van harte uitgenodigd om daar een aanpassing van
 *  of aanvulling op deze code voor te schrijven.
 *  Nieuwe features in deze software door mij komen er alleen als ik daar zelf behoefte aan heb.
 *  
 *  veranderingen voor versie 0.3:
 *  - leesbaarheid vergroten van de broncode
 *  - opschonen van de broncode
 *  - nog geen aansturing van de relais voor de aangesloten sporen! (= to-do voor een volgende versie)
 */

 // I2C library gebruiken:
 
#include <Wire.h>

// inputs
const int spoor1checkpin = 2;
const int linkspin = A3; // voor handbediening direct op de arduino, aan massa leggen
const int rechtspin = A2;// voor handbediening direct op de arduino, aan massa leggen
const int draaienfeedbackpin = 3;

// outputs
const int draairichtingpin = 4;
const int draaistartpin = 5;
const int spoor1 = 6;
const int spoor2 = 7;
const int spoor3 = 8;
const int spoor4 = 9;
const int spoor5 = 10;
const int spoor6 = 11;
const int spoor7 = 12;
const int bezigpin = 13;

// interne constanten
const int bezigvertraging = 10;
const int startvertraging = 400; // de tijd in mili seconden dat de electromagneet in de draaischijfmotor geactiveerd blijft
const int aantalsporen = 7;
const int linksom = 1;
const int rechtsom = 0;

// interne globale variabelen
int linksdraaien = 0;
int rechtsdraaien = 0;
int actiefspoor = 0;
int draairichting = 0;
int draaienfeedback = 0;
int spoor1check = 0;
int gewenstspoor = 1; // val terug op het eerste spoor
int aantalstappen = 0;
int state = 2; // 0 = ready, 1 = bezig, 2 = reset

void setup() {
  //start i2c op
  Wire.begin(0x28);             // wordt I2C slave met adres hex28
  Wire.onRequest(requestEvent); // subroutine voor aanvragen van de I2C master
  Wire.onReceive(receiveEvent); // subroutine voor verwerking van gegevens van de I2C master
  delay(2000); // geef de I2Cmaster de tijd om de reset op te merken
  
  // initialiseren inputs
  pinMode(linkspin, INPUT_PULLUP);
  pinMode(rechtspin, INPUT_PULLUP);
  pinMode(draaienfeedbackpin, INPUT_PULLUP);
  pinMode(spoor1checkpin, INPUT_PULLUP);
     
  // initialiseren outputs
  pinMode(draairichtingpin, OUTPUT);
  pinMode(draaistartpin, OUTPUT);
  pinMode(bezigpin, OUTPUT);

  spoorreset(); // draaibrug staat nu bij spoor 1 en actiefspoor = 1
}
// einde setup

// -----------------------------------------------------------------------------------------------------

// start main loop
void loop() {
  // lees knoppen voor handbediening
  linksdraaien = !digitalRead(linkspin);
  rechtsdraaien = !digitalRead(rechtspin);
  if (linksdraaien == HIGH) {
    if (actiefspoor > 1) {
      gewenstspoor = gewenstspoor - 1;
    }
  }  
  if (rechtsdraaien == HIGH) {
    if (actiefspoor < aantalsporen) {
      gewenstspoor = gewenstspoor + 1;
    }
  }
  spoorset(gewenstspoor);
}
// einde main loop

// ---------------------------------------------------------------------------------------------------------

// subroutines
void draaibrug(int draaibrugdraairichting, int draaibrugstappen) {
  int draaibrugdraaistart = 1;
  int draaibrugactiefspoor = actiefspoor;
  for (int i=1; i <= draaibrugstappen; ++i) {
    if (draaibrugdraairichting == linksom) {
      draaibrugactiefspoor = draaibrugactiefspoor - 1;
      if (draaibrugactiefspoor <= 1) i = draaibrugstappen; // begrenzing op laagst beschikbare spoor
    }
    if (draaibrugdraairichting == rechtsom) {
      draaibrugactiefspoor = draaibrugactiefspoor + 1;
      if (actiefspoor >= aantalsporen) i = draaibrugstappen; // begrenzing op hoogst beschikbare spoor
    } 
    state = 1;
    digitalWrite(bezigpin, 1);
    digitalWrite(draairichtingpin, draaibrugdraairichting);
    digitalWrite(draaistartpin, 1);
    delay(startvertraging);
    if (i == draaibrugstappen) draaibrugdraaistart = 0;
    digitalWrite(draaistartpin, 0);
    draaienfeedback = !digitalRead(draaienfeedbackpin);
    while (draaienfeedback) {
      digitalWrite(draairichtingpin, draaibrugdraairichting);
      delay(bezigvertraging);
      draaienfeedback = !digitalRead(draaienfeedbackpin);
    }
    actiefspoor = draaibrugactiefspoor;
  }
  digitalWrite(draairichtingpin, 0); //relais uitschakelen voor ruststand
  digitalWrite(bezigpin, 0);
  state = 0;
}

void spoorreset() {
  // staan we op spoor1 ?
  spoor1check = !digitalRead(spoor1checkpin);
  if (spoor1check == HIGH) {
    // ja
    actiefspoor = 1;
    state = 0;
  }
  else { 
    // nee, dus stap voor stap linksom draaien tot we op spoor 1 staan
    while (spoor1check == LOW) {
      draaibrug(linksom,1);
      spoor1check = !digitalRead(spoor1checkpin);
    }
  }
  actiefspoor = 1;
}

void spoorset(int spoorsetgewenstspoor) {
  int spoorsetdraairichting = linksom;
  int spoorsetaantalstappen = abs(actiefspoor - spoorsetgewenstspoor);
  if (actiefspoor < spoorsetgewenstspoor) spoorsetdraairichting = rechtsom;
  if (spoorsetaantalstappen > 0) draaibrug(spoorsetdraairichting, spoorsetaantalstappen);
}

void requestEvent() {
  // stuur de status, gewenstspoor en actiefspoor naar de I2C master (3 x een int variabele)
  Wire.write(state);
  Wire.write(gewenstspoor);
  Wire.write(actiefspoor);
}

void receiveEvent(){
  // hoor van de I2C master naar welk spoor er gedraaid dient te worden (1x een int variabele)
  gewenstspoor = Wire.read();
  if (gewenstspoor >= aantalsporen) gewenstspoor = aantalsporen; // begrenzing op hoogst beschikbare spoor
  if (gewenstspoor <= 1 )gewenstspoor = 1; // begrenzing op laagst beschikbare spoor
}
